/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"

void CL_FinishTimeDemo (void);

/*
===============
CL_StopPlayback
===============
*/
void CL_StopPlayback (void)
{
	if (!cls.demoplayback)
		return;

	fclose (cls.demofile);

	cls.demoplayback	= false;
	cls.demofile		= NULL;
	cls.state			= ca_disconnected;

	if (cls.timedemo)
		CL_FinishTimeDemo ();
}

/*
===================
CL_WriteDemoMessage
===================
*/
void CL_WriteDemoMessage (void)
{
	int	len = LittleLong (net_message.cursize);

	fwrite (&len, 4, 1, cls.demofile);
	fwrite (net_message.data, net_message.cursize, 1, cls.demofile);
	fflush (cls.demofile);
}

/*
=============
CL_GetMessage
=============
*/
int CL_GetMessage (void)
{
	int	r = 0;
	
	if (cls.demoplayback)
	{
		if (cls.signon == SIGNONS)
		{
			if (cls.timedemo)
			{
				if (host_framecount == cls.td_lastframe)
					return 0;

				cls.td_lastframe = host_framecount;

				if (host_framecount == cls.td_startframe + 1)
					cls.td_starttime = realtime;
			}
			else if (cl.time <= cl.mtime[0])
				return 0;
		}
		
		fread (&net_message.cursize, 4, 1, cls.demofile);

		net_message.cursize = LittleLong (net_message.cursize);
		if (net_message.cursize > MAX_MSGLEN)
			Sys_Error ("Demo message > MAX_MSGLEN");
		r = fread (net_message.data, net_message.cursize, 1, cls.demofile);
		if (r != 1)
		{
			CL_StopPlayback ();
			return 0;
		}
	
		return 1;
	}

	while (1)
	{
		r = NET_GetMessage (cls.netcon);
		
		if (r != 1 && r != 2)
			return r;
	
		if (net_message.cursize == 1 && net_message.data[0] == svc_nop)
			Con_Printf ("<-- server to client keepalive\n");
		else
			break;
	}

	if (cls.demorecording)
		CL_WriteDemoMessage ();
	
	return r;
}

/*
=========
CL_Stop_f
=========
*/
void CL_Stop_f (void)
{
	if (cmd_source != src_command)
		return;

	if (!cls.demorecording)
	{
		Con_Printf ("Not recording a demo.\n");
		return;
	}

	SZ_Clear (&net_message);
	MSG_WriteByte (&net_message, svc_disconnect);
	CL_WriteDemoMessage ();

	fclose (cls.demofile);
	cls.demofile = NULL;
	cls.demorecording = false;
	Con_Printf ("Completed demo\n");
}

/*
===========
CL_Record_f
===========
*/
void CL_Record_f (void)
{
	int		c;
	char	name[MAX_OSPATH];

	if (cmd_source != src_command)
		return;

	c = Cmd_Argc();

	if (c != 2 && c != 3)
	{
		Con_Printf ("record <demoname> <map>\n");
		return;
	}

	if (strstr(Cmd_Argv(1), ".."))
	{
		Con_Printf ("Relative pathnames are not allowed.\n");
		return;
	}

	if (c == 2 && cls.state == ca_connected)
	{
		Con_Printf("Can not record - already connected to server\nClient demo recording must be started before connecting\n");
		return;
	}

	_snprintf (name, sizeof(name), "data/%s.dem", Cmd_Argv(1));
	
	if (c > 2)
		Cmd_ExecuteString ( va("map %s", Cmd_Argv(2)), src_command);
	
	Con_Printf ("recording to %s.\n", name);
	cls.demofile = fopen (name, "wb");
	if (!cls.demofile)
	{
		Con_Printf ("ERROR: couldn't open.\n");
		return;
	}

	cls.demorecording = true;
}

/*
=============
CL_PlayDemo_f
=============
*/
void CL_PlayDemo_f (void)
{
	char	name[256];

	if (cmd_source != src_command)
		return;

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("play <demoname> : plays a demo\n");
		return;
	}

	CL_Disconnect ();
	
	_snprintf (name, sizeof(name), "data/%s.dem", Cmd_Argv(1));

	Con_Printf ("Playing demo from %s.\n", name);

	cls.demofile = fopen(name,"rb");
	if (!cls.demofile)
	{
		Con_Printf ("ERROR: couldn't open.\n");
		cls.demonum = -1;
		return;
	}

	cls.demoplayback	= true;
	cls.state			= ca_connected;
}

/*
=================
CL_FinishTimeDemo
=================
*/
void CL_FinishTimeDemo (void)
{
	int		frames	= (host_framecount - cls.td_startframe) - 1;
	float	time	= realtime - cls.td_starttime;

	cls.timedemo = false;

	if (!time)
		time = 1;

	Con_Printf ("%i frames %5.3f seconds %5.3f fps\n", frames, time, frames/time);
}

/*
=============
CL_TimeDemo_f
=============
*/
void CL_TimeDemo_f (void)
{
	if (cmd_source != src_command)
		return;

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("timedemo <demoname> : gets demo speeds\n");
		return;
	}

	CL_PlayDemo_f ();
		
	cls.timedemo		= true;
	cls.td_startframe	= host_framecount;
	cls.td_lastframe	= -1;
}
