/*
Copyright (C) 2001-2002 Tomas Jakobsson.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../game.h"

miniracer_t	miniracer;

gamefuncs_t gamefuncs[NUM_ENGINE_FUNCS] =
{
	{WorldSpawn,	"worldspawn"},
	{LapTrigger1,	"lap_trigger1"},
	{LapTrigger2,	"lap_trigger2"},
	{LapTrigger3,	"lap_trigger3"},
	{WayPoint,		"waypoint"},
	{StartPoint,	"startpoint"},
	{FinishPoint,	"finishpoint"},
	{SUB_Null,		"light"}
};

void Setmodel (edict_t *ent, char *model)
{
	char	**check;
	model_t	*mod;
	int		i;

	for (i=0, check = sv.model_precache ; *check ; i++, check++)
		if (!MR_strcmp (*check, model))
			break;
			
	if (!*check)
		Host_Error ("no precache: %s\n", model);

	ent->v.model		= model;
	ent->v.modelindex	= i;

	mod = sv.models[(int)ent->v.modelindex];

	if (mod)
		SetMinMaxSize (ent, mod->mins, mod->maxs);
	else
		SetMinMaxSize (ent, vec3_origin, vec3_origin);
}

void SetMinMaxSize (edict_t *ent, float *min, float *max)
{
	int	i;
	
	for (i=0 ; i<3 ; i++)
		if (min[i] > max[i])
			Host_Error ("backwards mins/maxs");

	VectorCopy (min, ent->game.mins);
	VectorCopy (max, ent->game.maxs);

	VectorSubtract (ent->game.maxs, ent->game.mins, ent->game.size);
	
	SV_LinkEdict (ent, false);
}

edict_t *Find (edict_t *edict, char *s)
{
	int		e;	
	edict_t	*ed;

	e = NUM_FOR_EDICT (edict);
	if (!s)
		Host_Error ("Find: bad search string");
		
	for (e++ ; e < sv.num_edicts ; e++)
	{
		ed = EDICT_NUM(e);
		if (ed->free)
			continue;

		if (!ed->game.classname)
			continue;

		if (!MR_strcmp (ed->game.classname ,s))
			return ed;
	}

	return NULL;
}

edict_t *Nextent (edict_t *edict)
{
	int		e;
	edict_t	*ent;
	
	e = NUM_FOR_EDICT (edict);

	while (1)
	{
		e++;
		if (e == sv.num_edicts)
		{
			return sv.edicts;
		}

		ent = EDICT_NUM(e);

		if (!ent->free)
		{
			return ent;
		}
	}
}

edict_t *Spawn (void)
{
	edict_t	*ed;
	ed = ED_Alloc();
	return ed;
}

void PrecacheModel (char *model)
{
	int	i;
	
	if (sv.state != ss_loading)
		Host_Error ("PrecacheModel: Can only be done in spawn functions");

	for (i=0 ; i<MAX_MODELS ; i++)
	{
		if (!sv.model_precache[i])
		{
			sv.model_precache[i] = model;
			sv.models[i] = Mod_ForName (model, true);
			return;
		}
		if (!MR_strcmp (sv.model_precache[i], model))
			return;
	}

	Host_Error ("PrecacheModel: overflow");
}

float Min3 (float value1, float value2, float value3)
{
	float f = value1;
	
	if (value2 < f)	f = value2;
	if (value3 < f)	f = value3;

	return f;
}

trace_t Traceline (vec3_t v1, vec3_t v2, edict_t *ent)
{
	trace_t	trace;

	trace = SV_Move (v1, vec3_origin, vec3_origin, v2, true, ent);

	return trace;
}

float VectoYaw (vec3_t value)
{
	float	yaw;
	
	if (value[1] == 0 && value[0] == 0)
		yaw = 0;
	else
	{
		yaw = (int) RAD2DEG(MR_atan2 (value[1], value[0]));
		if (yaw < 0)
			yaw += 360;
	}

	return yaw;
}

void CalcYaw (edict_t *ent)
{
	float	ideal	= ent->game.ideal_yaw;
	float	current	= ANGLEMOD (ent->v.angles[1]);
	float	move	= ideal - current;

	if (move == 0)
		return;

	if (move < 5 && move > -5)
	{
		if (ent->game.yaw_speed > 0)
		{
			ent->game.yaw_speed -= 600 * host_frametime;
			if (ent->game.yaw_speed < 0) 
				ent->game.yaw_speed = 0;
		}
		if (ent->game.yaw_speed < 0)
		{
			ent->game.yaw_speed += 600 * host_frametime;
			if (ent->game.yaw_speed > 0) 
				ent->game.yaw_speed = 0;
		}
		return;
	}

	if (ideal > current)
	{
		if (move >= 180)
			move -= 360;
	}
	else
	{
		if (move <= -180)
			move +=  360;
	}

	if (move > 0)
	{
		ent->game.yaw_speed += 800 * host_frametime;
	}
	else
	{
		ent->game.yaw_speed -= 800 * host_frametime;
	}
}

void ChangeAngles (edict_t *ent)
{
	float mult;

	mult = Length(ent->game.velocity) / 50;

	if (mult > 1) 
		mult = 1;

	if (ent->game.yaw_speed > 200) 
		ent->game.yaw_speed = 200;

	if (ent->game.yaw_speed < -200) 
		ent->game.yaw_speed = -200;

	ent->v.angles[1] += ent->game.yaw_speed * host_frametime * mult;
}

void VectoAngles (vec3_t value, vec3_t newvalue)
{
	float	forward;
	float	yaw, pitch;
	
	if (value[1] == 0 && value[0] == 0)
	{
		yaw = 0;
		if (value[2] > 0)
			pitch = 90;
		else
			pitch = 270;
	}
	else
	{
		yaw = (int) RAD2DEG(MR_atan2 (value[1], value[0]));
		if (yaw < 0)
			yaw += 360;

		forward = MR_sqrt (value[0]*value[0] + value[1]*value[1]);
		pitch = (int) RAD2DEG(MR_atan2 (value[2], forward));
		if (pitch < 0)
			pitch += 360;
	}

	newvalue[0] = pitch;
	newvalue[1] = yaw;
	newvalue[2] = 0;
}

void Normalize (vec3_t value, vec3_t newvalue)
{
	float	length;
	
	length = value[0] * value[0] + value[1] * value[1] + value[2]*value[2];
	
	if (length == 0)
		newvalue[0] = newvalue[1] = newvalue[2] = 0;
	else
	{
		length = MR_sqrt (length);
		length = 1/length;
		newvalue[0] = value[0] * length;
		newvalue[1] = value[1] * length;
		newvalue[2] = value[2] * length;
	}	
}

void SUB_Null (void)
{
}

void WayPoint (void)
{
	edict_t	*ent;

	ent = miniracer.self;

	if (ent->v.number < 0 || ent->v.number > 63)
		Host_Error ("Invalid waypoint number: %f", ent->v.number);

	if (miniracer.waypoint[(int)ent->v.number].used == true)
		Host_Error ("Duplicate waypoint : %f", ent->v.number);

	VectorCopy (ent->v.origin, miniracer.waypoint[(int)ent->v.number].origin);
	miniracer.waypoint[(int)ent->v.number].used = true;

	ED_Free (ent);

	miniracer.TotalWays++;
}

void StartPoint (void)
{
	edict_t	*ent;

	ent = miniracer.self;

	if (ent->v.number < 0 || ent->v.number > 7)
		Host_Error ("Invalid startpoint number: %f", ent->v.number);

	VectorCopy (ent->v.origin, miniracer.startpoint[(int)ent->v.number].origin);
	VectorCopy (ent->v.angles, miniracer.startpoint[(int)ent->v.number].angles);
	miniracer.startpoint[(int)ent->v.number].used = true;

	ED_Free (ent);
}

void FinishPoint (void)
{
	edict_t	*ent;

	ent = miniracer.self;

	if (ent->v.number < 0 || ent->v.number > 7)
		Host_Error ("Invalid finishpoint number: %f", ent->v.number);
	
	VectorCopy (ent->v.origin, miniracer.finishpoint[(int)ent->v.number].origin);
	VectorCopy (ent->v.angles, miniracer.finishpoint[(int)ent->v.number].angles);
	miniracer.finishpoint[(int)ent->v.number].used = true;

	ED_Free (ent);
}

void RestartTimer (void)
{
	static	float	timer;
	edict_t			*ent;
	char			mapname[32];
	int				num;

	ent = miniracer.self;
	ent->game.nextthink = sv.time + 5;
	ent->game.think		= RestartTimer;

	if (!miniracer.started)
		return;

	ent = Nextent (miniracer.world);
	num = 0;

	while (num < miniracer.MaxClients)
	{
		if (ent->game.Active && !ent->game.finished)
		{
			timer = sv.time + 20;
			return;
		}

		num += 1;
		ent = Nextent(ent);
	}	

	if (timer < sv.time)
	{
		ent->game.think		= SUB_Null;
		miniracer.started	= false;
		MR_strcpy (mapname, sv.name);
		SV_SpawnServer (mapname);
	}
}

void CarTouch (void)
{
}
