/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// mathlib.c -- math primitives

#include	"../quakedef.h"
#include	<math.h>

vec3_t vec3_origin = {0,0,0};

int BoxOnPlaneSideX (vec3_t emins, vec3_t emaxs, mplane_t *p) {return p->dist <= emins[0] ? 1 : (p->dist >= emaxs[0] ? 2 : 3);}
int BoxOnPlaneSideY (vec3_t emins, vec3_t emaxs, mplane_t *p) {return p->dist <= emins[1] ? 1 : (p->dist >= emaxs[1] ? 2 : 3);}
int BoxOnPlaneSideZ (vec3_t emins, vec3_t emaxs, mplane_t *p) {return p->dist <= emins[2] ? 1 : (p->dist >= emaxs[2] ? 2 : 3);}
int BoxOnPlaneSide0 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2]) >= p->dist) | (((p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2]) < p->dist) << 1));}
int BoxOnPlaneSide1 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2]) >= p->dist) | (((p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2]) < p->dist) << 1));}
int BoxOnPlaneSide2 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2]) >= p->dist) | (((p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2]) < p->dist) << 1));}
int BoxOnPlaneSide3 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2]) >= p->dist) | (((p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2]) < p->dist) << 1));}
int BoxOnPlaneSide4 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2]) >= p->dist) | (((p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2]) < p->dist) << 1));}
int BoxOnPlaneSide5 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2]) >= p->dist) | (((p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2]) < p->dist) << 1));}
int BoxOnPlaneSide6 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2]) >= p->dist) | (((p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2]) < p->dist) << 1));}
int BoxOnPlaneSide7 (vec3_t emins, vec3_t emaxs, mplane_t *p) {return (((p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2]) >= p->dist) | (((p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2]) < p->dist) << 1));}

void Sys_Error (char *error, ...);

/*-----------------------------------------------------------------*/

void BoxOnPlaneSideClassify (mplane_t *p)
{
	switch (p->type)
	{
	case 0: // x axis
		p->BoxOnPlaneSideFunc = BoxOnPlaneSideX;
		break;
	case 1: // y axis
		p->BoxOnPlaneSideFunc = BoxOnPlaneSideY;
		break;
	case 2: // z axis
		p->BoxOnPlaneSideFunc = BoxOnPlaneSideZ;
		break;
	default:
		if (p->normal[2] < 0) // 4
		{
			if (p->normal[1] < 0) // 2
			{
				if (p->normal[0] < 0) // 1
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide7;
				else
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide6;
			}
			else
			{
				if (p->normal[0] < 0) // 1
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide5;
				else
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide4;
			}
		}
		else
		{
			if (p->normal[1] < 0) // 2
			{
				if (p->normal[0] < 0) // 1
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide3;
				else
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide2;
			}
			else
			{
				if (p->normal[0] < 0) // 1
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide1;
				else
					p->BoxOnPlaneSideFunc = BoxOnPlaneSide0;
			}
		}
		break;
	}
}

double MR_pow (double x, double y)
{
	return pow (x, y);
}

double MR_fabs (double x)
{
	return fabs (x);
}

double MR_floor (double x)
{
	return floor (x);
}

double MR_ceil (double x)
{
	return ceil (x);
}

static float sintable[1024];

void BuildSinTable (void) 
{
	int i;

	for (i = 0; i < 1024; i++)
		sintable[i] = (float)sin(i * 3.14159265358979323846f / 2048);
}

double MR_sin (double x)
{
#ifdef _NATIVE_MATH
	return sin(x);	
#else
	int	index	= (int)(1024 * x / 1.57079632679489661923f);
	int	quad	= index >> 10;

	index	&= 1023;
	quad	&= 3;

	switch (quad) 
	{
		case 0:	return  sintable[index];
		case 1:	return  sintable[1023-index];
		case 2:	return -sintable[index];
		case 3:	return -sintable[1023-index];
	}

	return 0;
#endif
}

double MR_cos (double x)
{
#ifdef _NATIVE_MATH
	return cos(x);
#else
	int	index	= (int)(1024 * x / 1.57079632679489661923f);
	int	quad	= index >> 10;

	index	&= 1023;
	quad	&= 3;

	switch (quad)
	{
		case 0:	return  sintable[1023-index];
		case 1:	return -sintable[index];
		case 2:	return -sintable[1023-index];
		case 3:	return  sintable[index];
	}

	return 0;
#endif
}

double MR_asin (double x)
{
#ifdef _NATIVE_MATH
	return asin(x);
#else
	double		degree;
	int			i;
	qboolean	inv = false;

	if (x < 0)
	{
		inv = true;
		x = -x;
	}

	for (i=0; i<1024; i++)
	{
		if (sintable[i] >= x)
			break;
	}
	
	degree = i * 3.14159265358979323846f / 2048;

	return inv ? -degree : degree;
#endif
}

double MR_atan2 (double x, double y)
{
#ifdef _NATIVE_MATH
	return atan2(x, y);
#else
	double  x1 = x, y1 = y;
	float	base = 0, dir = 1, temp, test;
	int		i;

	if (y1 < 0) 
	{
		if (x1 >= 0) 
		{
			base	= 1.57079632679489661923f;
			temp	= y1;
			y1		= x1;
			x1		= -temp;
		} 
		else 
		{
			base	= 3.14159265358979323846f;
			x1		= -x1;
			y1		= -y1;
		}
	} 
	else 
	{
		if (x1 < 0) 
		{
			base	= 4.71238898038468985769f;
			temp	= y1;
			y1		= -x1;
			x1		= temp;
		}
	}

	if (x1 > y1) 
	{
		base	+= 1.57079632679489661923f;
		temp	= y1;
		y1		= x1;
		x1		= temp;
		dir		= -1;
	} 

	if (y1 == 0)
	{
		return base;
	}

	x1 /= y1;

	for (i=0; i<512; i++)
	{
		test = sintable[i] / sintable[1023-i];
		if (test > x1)
			break;
	}

	return base + dir * i * (3.14159265358979323846f / 2048);
#endif
}

void AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up)
{
	float	sr, sp, sy, cr, cp, cy, angle, spcy, spsy;
	
	angle	= DEG2RAD (angles[0]);
	sp		= MR_sin (angle);
	cp		= MR_cos (angle);

	angle	= DEG2RAD (angles[1]);
	sy		= MR_sin (angle);
	cy		= MR_cos (angle);

	angle	= DEG2RAD (angles[2]);
	sr		= MR_sin (angle);
	cr		= MR_cos (angle);

	spcy = sp * cy;
	spsy = sp * sy;

	forward[0]	= cp * cy;
	forward[1]	= cp * sy;
	forward[2]	= -sp;

	right[0]	= ((-sr * spcy) + (-cr * -sy));
	right[1]	= ((-sr * spsy) + (-cr *  cy));
	right[2]	= -sr * cp;

	up[0]		= (( cr * spcy) + (-sr * -sy));
	up[1]		= (( cr * spsy) + (-sr *  cy));
	up[2]		=  cr*cp;
}

void VectorNormalize (vec3_t v)
{
	float length = DotProduct (v,v);

	if (length)
	{
		length = MR_sqrt (length);

		length = 1/length;

		VectorScale (v, length, v);
	}
}

#ifdef _MASM
float DotProduct (float v1[3], float v2[3])
{
	float dotret;

	__asm
	{
		mov				ecx, v1
		mov				eax, v2

		fld		dword	ptr	[eax+0]
		fmul	dword	ptr	[ecx+0]
		fld		dword	ptr	[eax+4]
		fmul	dword	ptr	[ecx+4]
		fld		dword	ptr	[eax+8]
		fmul	dword	ptr	[ecx+8]
		fxch            st(1)
		faddp           st(2),st(0)
		faddp           st(1),st(0)
		fstp	dword	ptr	[dotret]
	}

	return dotret;
}
#elif defined _NO_ASM
float DotProduct(float v1[3], float v2[3])
{
	return v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2];
}
#else
float DotProduct(float v1[3], float v2[3])
{
	float dotret;
	
	__asm__ __volatile__
	("mov	%1,%%ecx\t\n"
	"mov	%2,%%eax\t\n"
	"fld	0(%%eax)\t\n"	
	"fmul	0(%%ecx)\t\n"	
	"fld	4(%%eax)\t\n"	
	"fmul	4(%%ecx)\t\n"	
	"fld	8(%%eax)\t\n"	
	"fmul	8(%%ecx)\t\n"	
	"fxch	%%st(1)\t\n"
	"faddp	%%st(0),%%st(2)\t\n"
	"faddp	%%st(0),%%st(1)\t\n"
	"fstp	%0"
	:"=m"(dotret)
	:"r"(v1), "r"(v2)
	:"%eax", "%ecx");
	
	return dotret;
}
#endif


#ifdef _WIN32
#define FP_BITS(fp) (* (DWORD *) &(fp))
#else
#define FP_BITS(fp) (* (unsigned long *) &(fp))
#endif

typedef union FastSqrtUnion
{
	float			f;
	unsigned int	i;
} FastSqrtUnion;

static unsigned int	FastSqrtTable[0x10000];

void BuildSqrtTable (void)
{
	unsigned int	i;
	FastSqrtUnion	s;

	for (i = 0; i <= 0x7FFF; i++)
	{
		s.i = (i << 8) | (0x7F << 23);
		s.f = (float) sqrt(s.f);
    
		FastSqrtTable[i + 0x8000] = (s.i & 0x7FFFFF);
    
		s.i = (i << 8) | (0x80 << 23);
		s.f = (float) sqrt(s.f);
    
		FastSqrtTable[i] = (s.i & 0x7FFFFF);
	}
}

float MR_sqrt (float n)
{
#ifdef _NATIVE_MATH
	return sqrt(n);
#else
	if (FP_BITS (n) == 0)
		return 0.0;                 
  
	FP_BITS (n) = FastSqrtTable[(FP_BITS (n) >> 8) & 0xFFFF] | ((((FP_BITS (n) - 0x3F800000) >> 1) + 0x3F800000) & 0x7F800000);
  
	return n;
#endif
}

int MR_rand (void)
{
	return (rand ());
}

void Math_Init (void)
{
#ifndef _NATIVE_MATH
	BuildSqrtTable ();
	BuildSinTable ();
#endif
}

