/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include	"../quakedef.h"
#include	"../vid.h"

int					timesort[8];
int					timelines;

extern	qboolean	showtimes;

/*
=========
ShowTimes
=========
*/
void ShowTimes (void)
{
	showtimes = true;
}

/*
=============
DontShowTimes
=============
*/
void DontShowTimes (void)
{
	showtimes = false;
}

/*
==========
Times_Init
==========
*/
void Times_Init (void)
{
	Cmd_AddCommand ("+showtimes", ShowTimes);
	Cmd_AddCommand ("-showtimes", DontShowTimes);
}

/*
=========
DrawTimes
=========
*/
void DrawTimes (void)
{
	int				i, y, j, k, l;
	int				minutes;
	int				seconds;
	int				hundreds;
	times_t			*times;
	byte			red, green, blue;

	if (!showtimes)
		return;

	timelines = 0;
	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.times[i].name[0])
		{
			timesort[timelines] = i;
			timelines++;
		}
	}

	for (i=0 ; i<timelines ; i++)
	{
		for (j=0 ; j<timelines-1-i ; j++)
		{
			if ((cl.times[timesort[j]].totaltime > cl.times[timesort[j+1]].totaltime))
			{
				k = timesort[j];
				timesort[j] = timesort[j+1];
				timesort[j+1] = k;
			}
		}
	}

	Draw_String (64,  16, "Name", 0, 255, 255, 255, 255);
	Draw_String (256, 16, "Fastest Lap", 0, 255, 255, 255, 255);
	Draw_String (448, 16, "Finish Time", 0, 255, 255, 255, 255);

// draw the text
	l = timelines;
	y = 48;
	for (i=0 ; i<l ; i++)
	{
		k = timesort[i];
		times = &cl.times[k];
		if (!times->name[0])
			continue;

		red = green = blue =255;

// draw name
		if (i == 0)
		{
			red		= 244;
			green	= 178;
			blue	= 0;
		}
		else if(i == 1)
		{
			red		= green = 209;
			blue	= 208;
		}
		else if (i == 2)
		{
			red		= 200;
			green	= blue = 0;
		}

		Draw_String (64, y, times->name, 0, red, green, blue, 255);

		if (times->fastestlap > 3000)
		{
			Draw_String (256, y, "--:--:--", 0, red, green, blue, 255);
			Draw_String (448, y, "--:--:--", 0, red, green, blue, 255);
		}
		else
		{
			minutes		= times->fastestlap / 60;
			seconds		= times->fastestlap - (60 * minutes);
			hundreds	= ((times->fastestlap - (60 * minutes)) - seconds) * 100;

			Draw_String (256, y, va("%02i:%02i:%02i",minutes, seconds, hundreds), 0, red, green, blue, 255);

			if (times->laprecord)
				Draw_String (172, y, "RECORD", 0, red, green, blue, 255);

			minutes		= times->totaltime / 60;
			seconds		= times->totaltime - (60 * minutes);
			hundreds	= (times->totaltime - (60 * minutes) - seconds) * 100;

			Draw_String (448, y, va("%02i:%02i:%02i",minutes, seconds, hundreds), 0, red, green, blue, 255);
			if (times->racerecord)
				Draw_String (364, y, "RECORD", 0, red, green, blue, 255);
		}

		y += 16;
	}
}

/*
========
DrawLaps
========
*/
void DrawLaps (void)
{
	int	y, i;

	y = vid.height - 128;

	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.times[i].name[0])
		{
			if (cl.times[i].on_lap != 4)
				Draw_String (0, y, va("%i", cl.times[i].on_lap), 0, cl.carcolor[i][0], cl.carcolor[i][1], cl.carcolor[i][2], 255);
			else
				Draw_String (0, y, "Finish", 0, cl.carcolor[i][0], cl.carcolor[i][1], cl.carcolor[i][2], 255);
			y += 16;
		}
	}
}
