/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"

#define	BLOCK_WIDTH		128
#define	BLOCK_HEIGHT	128
#define	MAX_LIGHTMAPS	64

int			lightmap_textures[MAX_LIGHTMAPS];
int			allocated[MAX_LIGHTMAPS][BLOCK_WIDTH];

byte		lightmaps[4*MAX_LIGHTMAPS*BLOCK_WIDTH*BLOCK_HEIGHT];

vec3_t		modelorg;

msurface_t  *skychain	= NULL;
msurface_t  *signchain	= NULL;
msurface_t  *waterchain	= NULL;

#ifdef _WIN32
extern void	(APIENTRY *MRglMTexCoord2f)	(GLenum, GLfloat, GLfloat);
extern void	(APIENTRY *MRglSelectTexture)	(GLenum);
#else
extern void	(*MRglMTexCoord2f)	(GLenum, GLfloat, GLfloat);
extern void 	(*MRglSelectTexture)	(GLenum);
#endif

extern	qboolean	gl_mtexable;

extern	GLenum		TEXTURE0;
extern	GLenum		TEXTURE1;

extern	camera_t	camera;

/*
===============
R_BuildLightMap
===============
*/
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride)
{
	int				smax, tmax;
	int				t;
	int				i, j, size;
	byte			*lightmap;
	unsigned		*bl;
	unsigned		blocklights[128*128*3];

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	size = smax*tmax;
	lightmap = surf->samples;

// set to full bright if no light data
	if (!cl.worldmodel->lightdata)
	{
		bl = blocklights;
		for (i=0 ; i<size ; i++)
		{
			bl[0]	= 255*256;
			bl[1]	= 255*256;
			bl[2]	= 255*256;
			bl		+= 3;
		}
		goto store;
	}

// clear to no light

	bl = blocklights;
	for (i=0 ; i<size ; i++) 
	{
		bl[0]	= 0;
		bl[1]	= 0;
		bl[2]	= 0;
		bl		+= 3;
	}

// add all the lightmaps
	if (lightmap)
	{
		bl = blocklights;

		for (i=0 ; i<size ; i++)
		{
			bl[0]		= lightmap[0] * 256;
			bl[1]		= lightmap[1] * 256;
			bl[2]		= lightmap[2] * 256;

			bl			+= 3;
			lightmap	+= 3;
		}
	}

// bound, invert, and shift
store:
	stride -= (smax<<2);
	bl = blocklights;
	for (i=0 ; i<tmax ; i++, dest += stride)
	{
		for (j=0 ; j<smax ; j++)
		{
			t = bl[0] >> 7;if (t > 255) t = 255;dest[0] = t;
			t = bl[1] >> 7;if (t > 255) t = 255;dest[1] = t;
			t = bl[2] >> 7;if (t > 255) t = 255;dest[2] = t;
			dest[3] = 255;

			bl		+= 3;
			dest	+= 4;
		}
	}
}

/*
==============
R_DrawMTexPoly
==============
*/
void R_DrawMTexPoly (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;

	p = s->polys;
	v = p->verts[0];

	glBindTexture (GL_TEXTURE_2D, lightmap_textures[s->lightmaptexturenum]);

	if (gl_mtexable)
	{
	    p = s->polys;
	    v = p->verts[0];

	    glBindTexture (GL_TEXTURE_2D, lightmap_textures[s->lightmaptexturenum]);

	    glBegin (GL_POLYGON);
		for (i=0; i<p->numverts; i++, v+= VERTEXSIZE)
		{
			MRglMTexCoord2f	(TEXTURE0, v[3], v[4]);
			MRglMTexCoord2f	(TEXTURE1, v[5], v[6]);
			glVertex3fv		(v);
		}
	    glEnd ();
	}
	else
	{
	    p = s->polys;
	    v = p->verts[0];

	    glBegin(GL_POLYGON);
	    for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	    {                                                                                       
		glTexCoord2f (v[3], v[4]);
		glTexCoord2f (v[5], v[6]);
		glVertex3fv (v);                                                        
	    }                                                                               
	    glEnd ();                                                                                                                                                       
	    
	    glBindTexture (GL_TEXTURE_2D, lightmap_textures[s->lightmaptexturenum]);
	    glEnable (GL_BLEND);
	    glBegin (GL_POLYGON);
	    v = p->verts[0];
	    for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	    {
		glTexCoord2f (v[5], v[6]);
		glVertex3fv (v);
	    }
	    glEnd ();
	    
	    glDisable (GL_BLEND);
	}

}

/*
================
R_DrawNoMTexPoly
================
*/
void R_DrawNoMTexPoly (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;

	p = s->polys;
	v = p->verts[0];

	glBegin (GL_POLYGON);
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			glTexCoord2f	(v[3], v[4]);
			glVertex3fv		(v);
		}
	glEnd ();
}

/*
========================
R_DrawNoMTexPolyLightMap
========================
*/
void R_DrawNoMTexPolyLightMap (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;

	p = s->polys;
	v = p->verts[0];

	glBindTexture	(GL_TEXTURE_2D, lightmap_textures[s->lightmaptexturenum]);

	glBegin (GL_POLYGON);
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			glTexCoord2f	(v[5], v[6]);
			glVertex3fv		(v);
		}
	glEnd ();
}

/*
=================
DrawTextureChains
=================
*/
void DrawTextureChains (void)
{
	int		i;
	msurface_t	*s;
	texture_t	*t, *tx;

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];

		if (!t)
			continue;

		s = t->texturechain;

		if (!s)
			continue;

		tx = s->texinfo->texture;

		glBindTexture (GL_TEXTURE_2D, tx->gl_texturenum);

		if (gl_mtexable)
		{
			MRglSelectTexture	(TEXTURE1);
			glEnable		(GL_TEXTURE_2D);

			for (; s; s=s->texturechain)
				R_DrawMTexPoly (s);

			glDisable		(GL_TEXTURE_2D);
			MRglSelectTexture	(TEXTURE0);
		}
		else
		{
			for (; s; s=s->texturechain)
				R_DrawNoMTexPoly (s);

			s = t->texturechain;

			glEnable	(GL_BLEND);
			glBlendFunc	(GL_ZERO, GL_SRC_COLOR);

			for (; s; s=s->texturechain)
				R_DrawNoMTexPolyLightMap (s);

			glBlendFunc	(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable	(GL_BLEND);
		}

		t->texturechain = NULL;
	}
}

/*
================
R_DrawBrushModel
================
*/
void R_DrawBrushModel (entity_t *e)
{
	vec3_t		mins, maxs;
	int			i;
	msurface_t	*psurf;
	float		dot;
	mplane_t	*pplane;
	model_t		*clmodel;
	texture_t	*tx;
	entity_t	*entity;

	entity	= e;
	clmodel = e->model;

	VectorAdd (e->origin, clmodel->mins, mins);
	VectorAdd (e->origin, clmodel->maxs, maxs);

	if (R_CullBox (mins, maxs))
		return;

	VectorSubtract (camera.origin, e->origin, modelorg);

	psurf = &clmodel->surfaces[clmodel->firstmodelsurface];

    glPushMatrix ();

    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
    glRotatef (e->angles[0],  0, 1, 0);
    glRotatef (e->angles[2],  1, 0, 0);

	//
	// draw texture
	//
	for (i=0 ; i<clmodel->nummodelsurfaces ; i++, psurf++)
	{
	// find which side of the node we are on
		pplane = psurf->plane;

		dot = PlaneDiff (modelorg, pplane);

	// draw the polygon
		if (( (psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
			(!(psurf->flags & SURF_PLANEBACK) && (dot >  BACKFACE_EPSILON)))
		{
			tx = psurf->texinfo->texture;

			glBindTexture (GL_TEXTURE_2D, tx->gl_texturenum);
			
			if (gl_mtexable)
			{
				MRglSelectTexture (TEXTURE1);
				glEnable (GL_TEXTURE_2D);

				R_DrawMTexPoly (psurf);

				glDisable (GL_TEXTURE_2D);
				MRglSelectTexture (TEXTURE0);
			}
			else
			{
				R_DrawNoMTexPoly (psurf);
				R_DrawNoMTexPolyLightMap (psurf);
			}
		}
	}
	glPopMatrix ();
}

/*
=============================================================

	WORLD MODEL

=============================================================
*/

int	r_visframecount;
int	r_framecount;

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;

	if (node->contents == CONTENTS_SOLID)
		return;		// solid

	if (node->visframe != r_visframecount)
		return;

	if (R_CullBox (node->mins, node->maxs))
		return;
	
// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}
		return;
	}

// node is just a decision point, so go down the apropriate sides

// find which side of the node we are on
	plane	= node->plane;
	dot		= PlaneDiff (modelorg, plane);
	side	= dot < 0;

// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 -BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;
		{
			for ( ; c ; c--, surf++)
			{
				if (surf->visframe != r_framecount)
					continue;

				if ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK))
					continue;		// wrong side

				if (surf->flags & SURF_DRAWSKY) 
				{
					surf->texturechain = skychain;
					skychain = surf;
				}
				else if (surf->flags & SURF_DRAWSKYUP) 
				{
					surf->texturechain = skychain;
					skychain = surf;
				}
				else if (surf->flags & SURF_DRAWWATER) 
				{
					surf->texturechain = waterchain;
					waterchain = surf;
				}
				else if (surf->flags & SURF_DRAWSIGN) 
				{
					surf->texturechain = signchain;
					signchain = surf;
				}
				else
				{
					surf->texturechain = surf->texinfo->texture->texturechain;
					surf->texinfo->texture->texturechain = surf;
				}
			}
		}
	}

// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}

/*
===========
R_DrawWorld
===========
*/
void R_DrawWorld (void)
{
	extern	cvar_t	draw_sky;

	VectorCopy (camera.origin, modelorg);

	R_RecursiveWorldNode (cl.worldmodel->nodes);

	DrawTextureChains ();

	if (draw_sky.ivalue && skychain)
		R_DrawSkyBox ();

	if (signchain)
		R_DrawSign (signchain);

	skychain	= NULL;
	signchain	= NULL;	
}

byte	*vis;

/*
============
R_MarkLeaves
============
*/
void R_MarkLeaves (void)
{
	mnode_t	*node;
	int		i;

	if (r_oldviewleaf == r_viewleaf)
		return;
	
	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);
		
	for (i=0 ; i<cl.worldmodel->numleafs ; i++)
	{
		if (vis[i>>3] & (1<<(i&7)))
		{
			node = (mnode_t *)&cl.worldmodel->leafs[i+1];
			do
			{
				if (node->visframe == r_visframecount)
					break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}

/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
int AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum=0 ; texnum<MAX_LIGHTMAPS ; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i=0 ; i<BLOCK_WIDTH-w ; i++)
		{
			best2 = 0;

			for (j=0 ; j<w ; j++)
			{
				if (allocated[texnum][i+j] >= best)
					break;
				if (allocated[texnum][i+j] > best2)
					best2 = allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("AllocBlock: full");
	return 0;
}

mvertex_t	*r_pcurrentvertbase;
model_t		*currentmodel;

/*
=======================
BuildSurfaceDisplayList
=======================
*/
void BuildSurfaceDisplayList (msurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		*vec;
	float		s, t, offset_s=0, offset_t=0;
	glpoly_t	*poly;

// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = fa->numedges;
	vertpage = 0;

	//
	// draw texture
	//
	poly			= Hunk_AllocName (sizeof(glpoly_t) + (lnumverts-4) * VERTEXSIZE*sizeof(float), "unknown");
	poly->next		= fa->polys;
	poly->flags		= fa->flags;
	fa->polys		= poly;
	poly->numverts	= lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = r_pcurrentvertbase[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = r_pcurrentvertbase[r_pedge->v[1]].position;
		}

		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= fa->texinfo->texture->height;

		VectorCopy (vec, poly->verts[i]);

		poly->verts[i][3] = s;
		poly->verts[i][4] = t;

		//
		// lightmap texture coordinates
		//
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s -= fa->texturemins[0];
		s += fa->light_s*16;
		s += 8;
		s /= BLOCK_WIDTH*16; //fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t -= fa->texturemins[1];
		t += fa->light_t*16;
		t += 8;
		t /= BLOCK_HEIGHT*16; //fa->texinfo->texture->height;

		poly->verts[i][5] = s;
		poly->verts[i][6] = t;

		poly->verts[i][3] += 200;
		poly->verts[i][4] += 200;
		
		if (poly->verts[i][3] > offset_s)
			offset_s = poly->verts[i][3];

		if (poly->verts[i][4] > offset_t)
			offset_t = poly->verts[i][4];
	}

	offset_s = (MR_ceil (offset_s)) - 1;
	offset_t = (MR_ceil (offset_t)) - 1;

	for (i=0 ; i<lnumverts ; i++)
	{
		poly->verts[i][3] -= offset_s;
		poly->verts[i][4] -= offset_t;
	}

	poly->numverts = lnumverts;
}

/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	int		smax, tmax;
	byte	*base;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;

	surf->lightmaptexturenum = AllocBlock (smax, tmax, &surf->light_s, &surf->light_t);
	base = lightmaps + surf->lightmaptexturenum*4*BLOCK_WIDTH*BLOCK_HEIGHT;
	base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * 4;
	R_BuildLightMap (surf, base, BLOCK_WIDTH*4);
}

/*
=================
GL_BuildLightmaps
=================
*/
void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;

	MR_memset (allocated, 0, sizeof(allocated));

	r_framecount = 1;		// no dlightcache

	if(!lightmap_textures[0])
		glGenTextures (MAX_LIGHTMAPS, lightmap_textures);

	for (j=1 ; j<MAX_MODELS ; j++)
	{
		m = cl.model_precache[j];
		if (!m)
			break;
		if (m->name[0] == '*')
			continue;
		r_pcurrentvertbase = m->vertexes;
		currentmodel = m;
		for (i=0 ; i<m->numsurfaces ; i++)
		{
			if (m->surfaces[i].flags & SURF_DRAWSKY || m->surfaces[i].flags & SURF_DRAWSKYUP)
				continue;
			
			GL_CreateSurfaceLightmap (m->surfaces + i);
			BuildSurfaceDisplayList	 (m->surfaces + i);
		}
	}

 	if (gl_mtexable)
 		MRglSelectTexture (TEXTURE1);

	//
	// upload all lightmaps that were filled
	//
	for (i=0 ; i<MAX_LIGHTMAPS ; i++)
	{
		if (!allocated[i][0])
			break;		// no more used

		glBindTexture	(GL_TEXTURE_2D, lightmap_textures[i]);
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D	(GL_TEXTURE_2D, 0, 3, BLOCK_WIDTH, BLOCK_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, lightmaps+i*BLOCK_WIDTH*BLOCK_HEIGHT*4);
	}

 	if (gl_mtexable)
		MRglSelectTexture (TEXTURE0);
}
