/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// server.h
#include <setjmp.h>

typedef struct
{
	int			maxclients;
	struct client_s	*clients;		// [maxclients]
	qboolean	changelevel_issued;	// cleared when at SV_SpawnServer
} server_static_t;

//=============================================================================

typedef enum {ss_loading, ss_active} server_state_t;

typedef struct
{
	server_state_t	state;	// some actions are only valid during load

	sizebuf_t		reliable_datagram;
	sizebuf_t		datagram;
	sizebuf_t		signon;

	edict_t			*edicts;

	struct model_s 	*worldmodel;
	struct model_s	*models[MAX_MODELS];

	double			time;
	double			lastchecktime;
	
	int				lastcheck;						// used by PF_checkclient
	int				num_edicts;
	int				max_edicts;

	char			name[64];						// map name
	char			modelname[64];					// maps/<name>.bsp, for model_precache[0]
	char			*model_precache[MAX_MODELS];	// NULL terminated

	byte			reliable_datagram_buf[MAX_DATAGRAM];
	byte			datagram_buf[MAX_DATAGRAM];
	byte			signon_buf[8192];

	qboolean		active;							// false if only a net client
} server_t;

#define	NUM_PING_TIMES		16
#define	NUM_SPAWN_PARMS		16

typedef struct client_s
{
	struct qsocket_s	*netconnection;

	sizebuf_t			message;

	edict_t				*edict;							// EDICT_NUM(clientnum+1)

	double				last_message;					// reliable messages must be sent
														// periodically
	float				ping_times[NUM_PING_TIMES];
	float				spawn_parms[NUM_SPAWN_PARMS];	// spawn parms are carried from level to level

	int					num_pings;						// ping_times[num_pings%NUM_PING_TIMES]	

	char				name[32];						// for printing to other people
	char				carmodel[32];

	byte				msgbuf[MAX_MSGLEN];
	byte				carcolors[3];

	qboolean			active;							// false = client is free
	qboolean			spawned;						// false = don't send datagrams
	qboolean			dropasap;						// has been told to go to another level
	qboolean			sendsignon;						// only valid before spawned
} client_t;

//=============================================================================

// edict->movetype values
#define	MOVETYPE_NONE			0		// never moves
#define	MOVETYPE_WALK			1		// gravity
#define	MOVETYPE_PUSH			2		// no clip to world, push and crush

// edict->solid values
#define	SOLID_NOT				0		// no interaction with other objects
#define	SOLID_TRIGGER			1		// touch on edge, but not blocking
#define	SOLID_BBOX				2		// touch on edge, block
#define	SOLID_BSP				3		// bsp clip, touch on edge, block

//============================================================================

extern	server_static_t	svs;				// persistant server info
extern	server_t		sv;					// local server
extern	client_t		*host_client;
extern	jmp_buf 		host_abortserver;
extern	double			host_time;
extern	edict_t			*sv_player;

//===========================================================

int		SV_ModelIndex				(char *name);
void	SV_Init						(void);
void	SV_Physics					(void);
void	SV_RunClients				(void);
void	SV_DropClient				(qboolean crash);
void	SV_AddUpdates				(void);
void	SV_SpawnServer				(char *server);
void	SV_ClientPrintf				(char *fmt, ...);
void	SV_ClearDatagram			(void);
void	SV_BroadcastPrintf			(char *fmt, ...);
void	SV_SendClientMessages		(void);
void	SV_CheckForNewClients		(void);
