/*
Copyright (C) 2003 Martin Weiss.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef SOUND

#include	<SDL.h>
#include	<SDL_mixer.h>
#include	"soundfx.h"
#include	"../quakedef.h"

#define		MR_ENGINE_SOUND	"f40.wav"

extern	qboolean engine_started;

typedef struct
{
    Mix_Chunk *chunk;
    int channel;
    Uint8 *wavdata;
} soundbuf_t;

soundbuf_t soundbuf[32];

float shift=1.0;
int pitch_finished=1;

static int engine_soundnum=-1;

int audio_rate=22500; //11025;
Uint16 audio_format=AUDIO_S16;
int audio_channels=2; //1;	//we need stereo for fxShift()
int audio_buffers=4096;

static qboolean audio_avail=false;

//postmix callback function for engine sound pitch
void EnginePitch(void *data, Uint8 *stream, int len)
{
	int i, srclen;
	Uint8 srcptr[65536];
	static int pos=0;

	srclen = (int)((len / 4) * shift + 1) * 4;	//from gltron-0.61/sound3d.c
	for(i=0;i<srclen;i++)
	{
		if(pos == soundbuf[engine_soundnum].chunk->alen)
		    pos=0;
		srcptr[i]=soundbuf[engine_soundnum].chunk->abuf[pos];
		pos++;
	}
	
	//fxShift() does shift and mixes shifted sound to stream
	fxShift(shift, stream, srcptr, len);
}

qboolean MRS_Init (void)
{
	int i;
	
	if (COM_CheckParm("-nosound"))
	{
		Con_Printf("Sound disabled.\n");
		return false;
	}

	// initialize SDL audio
	if ( SDL_Init(SDL_INIT_AUDIO) < 0 )
	{
		Con_Printf("Couldn't initialize SDL: %s\n",SDL_GetError());
		return false;
	}
	
	// open the audio device
	if (Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers))
	{
		Con_Printf("Couldn't open audio: %s\n", SDL_GetError());
		return false;
	}
	
	// initialize the soundchannel array
	for(i=0;i<64;i++)
	{
		soundbuf[i].channel=-1;
	}
	
	audio_avail=true;
	return true;
}

void MRS_Shutdown (void)
{
	if(audio_avail)
	{
		Mix_SetPostMix(NULL, NULL);
		Mix_CloseAudio();
		SDL_Quit();
	}
}

qboolean MRS_SampleLoad (int number, char *filename)
{
	if(audio_avail)
	{
		soundbuf[number].wavdata=(Uint8 *)COM_LoadHunkFile(filename);
		soundbuf[number].chunk=Mix_QuickLoad_WAV(soundbuf[number].wavdata);
		//find the buffer with the engine sound
		if(strstr(filename, MR_ENGINE_SOUND) != NULL)
		{
			engine_soundnum=number;
		}

		return true;
	}
	else
		return false;
}

void MRS_SampleFree (int number)
{
	if(audio_avail)
	{
		//Mix_FreeChunk(soundbuf[number].chunk);
		soundbuf[number].chunk->allocated=0;
		if(number==engine_soundnum)
			engine_soundnum=-1;
	}
}

void MRS_SamplePlay (int number, int volume, qboolean loop)
{
	//static int pitchreg=0;
	
	if(audio_avail)
	{
		volume=volume/39+128;
		soundbuf[number].chunk->volume=volume;
		
		//register the function for engine sound
		if((number==engine_soundnum)) // && (pitchreg==0))
		{
			Mix_SetPostMix(EnginePitch, NULL);
			return;
			//pitchreg=1;
		}
		
		if(loop)
			soundbuf[number].channel=Mix_PlayChannel(-1, soundbuf[number].chunk, -1);
		else
			soundbuf[number].channel=Mix_PlayChannel(-1, soundbuf[number].chunk, 0);	
	}
}

void MRS_SampleStop (int number)
{
	if(audio_avail)
	{
		//unregister the function for engine sound
		if(number==engine_soundnum)
		{
			Mix_SetPostMix(NULL, NULL);
		}

		if(soundbuf[number].channel != -1)
		{
			Mix_HaltChannel(soundbuf[number].channel);
			soundbuf[number].channel=-1;
		}
	}
}

void MRS_SampleSetVolume (int number, int volume)
{
	if(audio_avail)
	{
		volume=volume/39+128;	//directx: -5000...0 -> sdl: 0...128
		soundbuf[number].chunk->volume=volume;
		if(soundbuf[number].channel != -1)
			Mix_Volume(soundbuf[number].channel, volume);
	}
}

void MRS_SampleSetPan (int number, int pan)
{

}

void MRS_SampleSetPitch (int number, int pitch)
{
	static float oldpitch = 7500.0;
	float dpitch;
	
	if(audio_avail)
	{
		if(number != engine_soundnum)
			return;
		if(engine_started)
		{
			// max. step width for pitching: 300Hz 
			dpitch = pitch - oldpitch;
			if(dpitch > 300.0)
			    pitch = oldpitch + 300.0;
			// max. step width for pitching at brake: 3000Hz
			else if(dpitch < -6000.0)
			    pitch = oldpitch - 3000.0;
			else if(dpitch < -300.0)
			    pitch = oldpitch - 300.0;
			
			shift=(float)pitch/11025.0;	//when pitch=11025(Hz) then shift=1
			
			oldpitch = pitch;
		}
	}
}

int MRS_SampleGetPitch (int number)
{
	return 0;
}

#endif
